import numpy as np
from abc import ABC, abstractmethod

from rafael_nn.common import FloatArr

class ActivationFunction(ABC):
    def __init__(self, n:int):
        self.n = n

    @abstractmethod
    def __call__(self,x:FloatArr) -> FloatArr:
        pass

    @abstractmethod
    def init_sample(self) -> np.float64:
        pass

# since the best initialization depends on the activation function. We delegate that to its class
class ReLU(ActivationFunction):
    def __call__(self,x:FloatArr):
        return x.clip(0)

    def init_sample(self) -> np.float64:
        return np.float64(np.random.normal(loc=0, scale=2 / self.n))




from typing import Any
import numpy as np
from abc import ABC, abstractmethod

from rafael_nn.common import FloatArr

class LossFunction(ABC):
    @abstractmethod
    def __call__(self, prediction: FloatArr, target: FloatArr) -> float:
        """Compute the loss value."""
        pass

    @abstractmethod
    def backward(self, prediction: FloatArr, target: FloatArr) -> FloatArr:
        """Compute the gradient of the loss with respect to the prediction."""
        pass

class MeanSquaredError(LossFunction):
    def __call__(self, prediction: FloatArr, target: FloatArr) -> np.floating[Any]:
        return np.mean((prediction - target) ** 2)

    def backward(self, prediction: FloatArr, target: FloatArr) -> FloatArr:
        return 2 * (prediction - target) / prediction.size





import unittest
import numpy as np
from rafael_nn.acfn import ReLU
from rafael_nn.layer import Layer, Linear
from rafael_nn.lossfn import MeanSquaredError

from rafael_nn.nn import NeuralNetwork
from rafael_nn.optimizer import GradientDescent

class TestNNGradient(unittest.TestCase):
    def setUp(self):
        np.random.seed(42)
        layers = 5
        n_by_layer = 6
        layers:list[Layer] = [Linear(1,n_by_layer)] + [Linear(n_by_layer,n_by_layer) for _ in range(layers-1)] + [Linear(n_by_layer,1)]

        self.loss_fn = MeanSquaredError()
        self.nn = NeuralNetwork(layers, optimizer=GradientDescent(), loss_fn=self.loss_fn)

    # def test_nn_gradient_with_bias_change(self):
    def test_nn_gradient_with_bias_change(self):
        """Here we perform the same kind of check made in the notebook 7_2 to see if our derivatives are well calculated"""
        x = np.array([[1.2]])
        target = np.array([[1.0]])
        epsilon = 0.000001 # same delta used in notebook 7_2

        prediction = self.nn(x)
        all_dl_b, _ = self.nn._backward(prediction, target)

        for layer_i in range(len(self.nn.layers)):
            layer = self.nn.layers[layer_i]

            bias_cpy = np.array(layer.biases)
            dl_bias = np.zeros_like(all_dl_b[layer_i])

            for row in range(len(layer.biases)):
                # get current output
                output = self.nn(x)
                layer.biases[row]+=epsilon
                # get output with small change
                output_change = self.nn(x)

                #restore previous value
                layer.biases = bias_cpy
                dl_bias[row] = (self.loss_fn(output_change,target) - self.loss_fn(output, target))/epsilon

            assert np.allclose(all_dl_b[layer_i], dl_bias, rtol=1e-5, atol=1e-8, equal_nan=False), \
                f"Gradient check failed for bias at layer {layer_i}.\n" \
                f"Backprop: {all_dl_b[layer_i]}\n" \
                f"Finite diff: {dl_bias}"

    def test_nn_gradient_with_weight_change(self):
        """Here we perform the same kind of check made in the notebook 7_2 to see if our derivatives are well calculated"""
        x = np.array([[1.2]])
        target = np.array([[1.0]])
        epsilon = 0.000001 # same delta used in notebook 7_2

        prediction = self.nn(x)
        _, all_dl_w = self.nn._backward(prediction, target)

        for layer_i in range(len(self.nn.layers)):
            layer = self.nn.layers[layer_i]

            weight_copy = np.array(layer.weights)
            dl_w = np.zeros_like(all_dl_w[layer_i])

            for row in range(layer.weights.shape[0]):
                for col in range(layer.weights.shape[1]):
                    # get current output
                    output = self.nn(x)
                    layer.weights[row][col]+=epsilon
                    # get output with small change
                    output_change = self.nn(x)

                    #restore previous value
                    layer.weights = weight_copy
                    dl_w[row][col] = (self.loss_fn(output_change,target) - self.loss_fn(output, target))/epsilon

            # NOTE: here I had to reduce atol from 1e-8 to 1e-5. Maybe some imprecision on my code?
            assert np.allclose(all_dl_w[layer_i], dl_w, rtol=1e-5, atol=1e-5, equal_nan=False), \
                f"Gradient check failed for weight at layer {layer_i}.\n" \
                f"Backprop: {all_dl_w[layer_i]}\n" \
                f"Finite diff: {dl_w}"





import unittest
import numpy as np
from rafael_nn.common import FloatArr

from rafael_nn.layer import Linear
from rafael_nn.lossfn import MeanSquaredError
from rafael_nn.nn import NeuralNetwork
from rafael_nn.optimizer import GradientDescent, StochasticGradientDescend

# This is the same toy function used in the example of backprop in understanding deep learning
def teaching_function(beta: np.ndarray, omega: np.ndarray):
    def fn_x_only(x:FloatArr) -> FloatArr:
        return beta[3] + omega[3] * np.cos(
            beta[2] + omega[2] * np.exp(
                beta[1] + omega[1] * np.sin(
                    beta[0] + omega[0] * x
                )
            )
        )
    return fn_x_only

def create_dataset(size:int, fn):
    np.random.seed(42)

    x = np.linspace(-5, 5, size)
    y_clean = fn(x)
    y_noisy = y_clean + np.random.normal(0, 0.1, size=x.shape)  # add small noise

    return x.reshape((1,size)), y_noisy.reshape((1,size))

class TestNNTrain(unittest.TestCase):
    def setUp(self):
        beta = np.random.uniform(-1, 1, size=4)
        omega = np.random.uniform(-1, 1, size=4)

        np.random.seed(42)
        n_layers = 2
        n_by_layer = 6
        layers= [Linear(1,n_by_layer)] + [Linear(n_by_layer,n_by_layer) for _ in range(n_layers-1)] + [Linear(n_by_layer,1)]

        self.teaching_fn = teaching_function(beta,omega)
        self.loss_fn = MeanSquaredError()
        self.nn = NeuralNetwork(layers, optimizer=StochasticGradientDescend(), loss_fn=self.loss_fn)

    def test_train_correct_descend(self):
        x_train, y_train = create_dataset(10, self.teaching_fn)
        self.nn.train(x_train,y_train)



from abc import ABC, abstractmethod
from typing import Optional
import numpy as np
from numpy.typing import NDArray # Import NDArray from numpy.typing

from rafael_nn.acfn import ActivationFunction, ReLU
from rafael_nn.common import FloatArr
from rafael_nn.optimizer import Optimizer

class Layer(ABC):
    weights: NDArray[np.float64]
    biases: NDArray[np.float64]
    fn: ActivationFunction

    f: FloatArr
    h: FloatArr

    @abstractmethod
    def __call__(self, input: FloatArr) -> tuple[FloatArr,FloatArr]:
        """Compute the loss value."""
        pass

    @abstractmethod
    def backward(self, prev_dl_f:Optional[FloatArr] = None, weights_dl_f: Optional[FloatArr] = None) -> tuple[FloatArr,FloatArr,FloatArr]:
        """Compute the gradient of the loss with respect to the prediction."""
        pass

    # for debugging, helped a lot when building backpropagation
    def __str__(self):
        return "Unknown"

    def __repr__(self):
        return "Unknown"

# NOTE: I had a design issue. Each layer was storing its preactivation and activation values.
# but for the first layer the activation value is the input. Because this should be the activation that happened on the PREVIOUS layer
# my design had these values stored on the layer they were executed, 
# so the first layer h value was the first execution of the activation function. And the last one didnt had one

# NOTE: to solve the issue above, just added 'h' as the input value the layer received when called on the forward pass

class Linear(Layer):
    prev:int
    neurons:int

    def __init__(self,prev:int, neurons:int, fn:Optional[ActivationFunction] = None):
        """Initializes linear layer with weights. Initializes biases with 0"""
        self.fn = ReLU(neurons) if fn is None else fn
        weights = [[self.fn.init_sample() for _ in range(prev)] for _ in range(neurons)]
        # TODO, output is 0 if bias is 0. need to check this
        #bug: before I was initializing biases with shape (6,). This was causing issues when added. It turned the vector into a matrix
        biases = [[0] for _ in range(neurons)]

        self.h = None
        self.f = None
        self.prev = prev
        self.neurons = neurons
        self.weights = np.array(weights, dtype=np.float64)
        self.biases = np.array(biases, dtype=np.float64)

    def __call__(self, input: FloatArr) -> tuple[FloatArr, FloatArr]:
        """Applies the forward pass. Multiplies the given vector by the matrix weights and applies the activation fn"""
        # we need the preactivation for the gradient calc
        self.h = input
        self.f = self.biases + self.weights@input

        return self.fn(self.f), self.f

    def backward(self, prev_dl_f:Optional[FloatArr] = None, weights_dl_f: Optional[FloatArr] = None) -> tuple[FloatArr,FloatArr,FloatArr]:
        """
        Calculates dl_weights, dl_f and dl_bias given arguments. weights_dl_f or prev_dl_f must be passed (one or the other).
        if prev_dl_f is passed, this means this is the LAST layer and the first to calculate dl_f. 
        if weights_dl_f is passed, it means this is a hidden layer
        """
        if prev_dl_f is None:
            dl_f = np.where(self.f > 0, 1, 0)*weights_dl_f
        else:
            dl_f = prev_dl_f

        dl_bias = np.sum(dl_f, axis=1).reshape(dl_f.shape[0], 1)

        # dl_bias, dl_f and dl_weight
        return dl_bias, dl_f, dl_f@self.h.T

    def __str__(self):
        return f"Linear({self.prev},{self.neurons})"

    def __repr__(self):
        return f"Linear({self.prev},{self.neurons})"




from .acfn import ActivationFunction, ReLU
from .common import FloatArr
from .layer import Layer,Linear
from .lossfn import LossFunction,MeanSquaredError
from .optimizer import Optimizer,GradientDescent, StochasticGradientDescend
from .nn import NeuralNetwork

__all__ = [
    "ActivationFunction",
    "ReLU",
    "FloatArr",
    "Layer",
    "Linear",
    "LossFunction",
    "MeanSquaredError",
    "Optimizer",
    "GradientDescent",
    "StochasticGradientDescend",
    "NeuralNetwork"
]



import numpy as np
import unittest
from unittest.mock import MagicMock
from rafael_nn.acfn import ActivationFunction
from rafael_nn.layer import Linear

class TestLinear(unittest.TestCase):

    def setUp(self):
        self.prev_neurons = 5
        self.current_neurons = 3
        # here I figured out how to make mocks for classes in python. Following best practices for unit tests
        # In unit tests, bugs in other classes should not affect the testing of the current one, thats why we mock
        self.mock_activation_fn = MagicMock(spec=ActivationFunction)
        self.mock_activation_fn.side_effect = lambda x: x + 1 # identity + 1 function for simplicity 
        self.mock_activation_fn.init_sample.side_effect = lambda: 0.1

        self.linear_layer = Linear(self.prev_neurons, self.current_neurons, self.mock_activation_fn)

    def test_initialization_weights_shape(self):
        "Here we test the shape is correct"
        expected_shape = (self.current_neurons, self.prev_neurons)
        self.assertEqual(self.linear_layer.weights.shape, expected_shape)

    def test_initialization_weights_dtype(self):
        "Here we test the numbers type is correct"
        self.assertEqual(self.linear_layer.weights.dtype, float)

    def test_initialization_biases_shape(self):
        "Here we test the biases shape is correct"
        expected_shape = (self.current_neurons,1)
        self.assertEqual(self.linear_layer.biases.shape, expected_shape)

    def test_initialization_fn_assignment(self):
        "Here we test the used activation function is correct"
        self.assertEqual(self.linear_layer.fn, self.mock_activation_fn)

    def test_initialization_weights_content(self):
        "Here we check the weights are correctly initialized"
        expected_weights = np.array([
            [0.1] * self.linear_layer.weights.shape[1],
            [0.1] * self.linear_layer.weights.shape[1],
            [0.1] * self.linear_layer.weights.shape[1],
        ], dtype=float)
        np.testing.assert_array_equal(self.linear_layer.weights, expected_weights)
        # in a single call should affect all input
        self.assertEqual(self.mock_activation_fn.init_sample.call_count, self.current_neurons * self.prev_neurons)

    def test_initialization_biases_content(self):
        "Here we check the biases are correctly initialized"
        expected_biases = np.array([[0]] * self.linear_layer.weights.shape[0], dtype=float)
        np.testing.assert_array_equal(self.linear_layer.biases, expected_biases)

    def test_forward_pass_output_shape(self):
        "Here we check the shape of the forward pass is correct"
        # test input vector
        input_vector = np.random.normal(size=(self.prev_neurons,1))
        output,_ = self.linear_layer(input_vector)
        self.assertEqual(output.shape, (self.current_neurons,1))

    def test_forward_pass_multiplication_and_activation(self):
        """Test that the forward pass correctly performs multiplication and applies activation."""

        test_input = np.array([[1.0], [2.0], [3.0], [4.0], [5.0]], dtype=float)
        # we multiply [1...5] by [[[0.1] * 5] * 3] and add 1 to each entry
        expected_raw_output = np.array([[1.5], [1.5], [1.5]], dtype=float)

        # here our mock activation fn is x + 1
        expected_final_output = expected_raw_output + 1
        output,_ = self.linear_layer(test_input)

        np.testing.assert_array_almost_equal(output, expected_final_output)
        self.mock_activation_fn.assert_called_once()
        np.testing.assert_array_almost_equal(self.mock_activation_fn.call_args[0][0], expected_raw_output)

# To run the tests:
if __name__ == '__main__':
    unittest.main()



import unittest
import numpy as np
from rafael_nn.acfn import ReLU

class TestReLU(unittest.TestCase):

    def test_forward_positive(self):
        relu = ReLU(1)
        x = np.array([1, -1, 0, 5])
        output = relu(x)
        expected = np.array([1, 0, 0, 5])
        np.testing.assert_array_equal(output, expected)

if __name__ == '__main__':
    unittest.main()




import numpy as np
from numpy.typing import NDArray
from rafael_nn.common import FloatArr

from rafael_nn.layer import Layer
from rafael_nn.lossfn import LossFunction
from rafael_nn.optimizer import Optimizer

np.random.seed(42)

class NeuralNetwork:
    # I like adding types. Its easier to know that what im doing will work, also easier to debug
    # want to update this to use functional programming
    layers: list[Layer]
    optimizer:Optimizer
    layers_output: list[NDArray[np.float64]]
    loss_fn:LossFunction

    def __init__(self, layers:list[Layer], optimizer:Optimizer, loss_fn:LossFunction):
        self.layers = layers
        self.optimizer = optimizer
        self.loss_fn = loss_fn

    def __call__(self, x: np.ndarray) -> np.ndarray:
        return self._forward(x)[0]

    def train(self, x_full: FloatArr, y_full:FloatArr, epochs = 1000, err = 1e-4):
        cur_err = np.inf
        self.optimizer.load_data(x_full,y_full)

        while epochs > 0 and cur_err > err:
            x, y = self.optimizer.get_batch()

            final = self(x)
            cur_err = self.loss_fn(final,y)
            all_dl_bias,all_dl_weights = self._backward(final, y)

            for i in range(len(self.layers)):
                layer = self.layers[i]
                layer.biases = self.optimizer(layer.biases,all_dl_bias[i])
                layer.weights = self.optimizer(layer.weights,all_dl_weights[i])
            epochs-=1

    def backward(self, prediction:FloatArr, target:FloatArr) -> tuple[list[FloatArr],list[FloatArr]]:
        return self._backward(prediction,target)

    # this is almos the same implementation as the 7_2 notebook
    def _forward(self, x: FloatArr) -> tuple[FloatArr, list[FloatArr], list[FloatArr]]:
        all_h, all_f = [], []
        # all layers but the last one apply activation fn
        for i in range(len(self.layers) - 1):
            layer = self.layers[i]
            h, f = layer(x if i == 0 else all_h[i-1])

            all_h.append(h)
            all_f.append(f)

        # last layer outputs result y
        _,res = self.layers[-1](all_h[-1])
        all_f.append(res)

        return res, all_h, all_f

    # THIS IS AN IMPLEMENTATION OF GRADIENT DESCEND
    def _backward(self, prediction:FloatArr, target:FloatArr) -> tuple[list[FloatArr],list[FloatArr]]:
        layers_n = len(self.layers)
        all_dl_bias, all_dl_weights = [np.array([])] * layers_n, [np.array([])] * layers_n

        dl_b, prev_dl_f, dl_w = self.layers[-1].backward(prev_dl_f=self.loss_fn.backward(prediction,target))
        all_dl_bias[-1] = dl_b
        all_dl_weights[-1] = dl_w

        for lay in range(layers_n - 2,-1,-1):
            # here had an anoying error because prev_w was already transposed on this line
            prev_w = self.layers[lay+1].weights
            dl_b, prev_dl_f, dl_w = self.layers[lay].backward(weights_dl_f=prev_w.T @ prev_dl_f)

            all_dl_bias[lay] = dl_b
            all_dl_weights[lay] = dl_w

        return all_dl_bias, all_dl_weights





import numpy as np
import unittest
from unittest.mock import MagicMock, patch
from rafael_nn.acfn import ActivationFunction
from rafael_nn.common import FloatArr
from rafael_nn.layer import Layer
from rafael_nn.lossfn import MeanSquaredError

from rafael_nn.nn import NeuralNetwork

class MockActFunction(ActivationFunction):
    """Mock for ActivationFunction."""
    def __call__(self, x: FloatArr) -> FloatArr:
        return x
    def init_sample(self, i: int, j: int) -> float:
        return 0.0

class TestLinear(Layer):
    """Mock for the Linear layer."""
    def __init__(self, prev: int, neurons: int, fn: MockActFunction):
        self.prev = prev
        self.neurons = neurons
        self.fn = fn
        self.weights = np.zeros((neurons, prev)) 

    def backward(self):
        pass

    def __call__(self, input: FloatArr):
        return input + 1, input + 1

class Optimizer:
    """Mock for the Optimizer."""
    def __init__(self):
        pass 
    def optimize(self, params):
        pass

class TestNeuralNetwork(unittest.TestCase):

    def setUp(self):
        """Set up common variables and mocks for tests."""
        self.mock_layer1 = TestLinear(0,0,MockActFunction(1))
        self.mock_layer2 = TestLinear(0,0,MockActFunction(1))
        self.mock_layer3 = TestLinear(0,0,MockActFunction(1))

        self.mock_optimizer = MagicMock(spec=Optimizer)

        self.neural_network = NeuralNetwork(
            layers=[self.mock_layer1, self.mock_layer2, self.mock_layer3],
            optimizer=self.mock_optimizer,
            loss_fn=MeanSquaredError()
        )

    def test_forward_pass_flow_and_output(self):
        initial_input = np.array([5.0, 5.0], dtype=np.float64)

        expected_l1 = initial_input + 1.0
        expected_l2 = expected_l1 + 1.0
        expected_l3 = expected_l2 + 1.0

        output = self.neural_network(initial_input)

        np.testing.assert_array_equal(output, expected_l3)

    def test_optimizer_assignment(self):
        """Test that the optimizer is correctly assigned."""
        self.assertEqual(self.neural_network.optimizer, self.mock_optimizer)

# To run the tests:
if __name__ == '__main__':
    unittest.main(argv=[''], exit=False)




import numpy as np
from abc import ABC, abstractmethod

from rafael_nn.common import FloatArr

class Optimizer(ABC):
    @abstractmethod
    def __call__(self, parameters: FloatArr, gradients: FloatArr) -> FloatArr:
        pass

    @abstractmethod
    def load_data(self, x_full: FloatArr, y_full:FloatArr):
        pass

    @abstractmethod
    def get_batch(self) -> tuple[FloatArr, FloatArr]:
        pass


class GradientDescent(Optimizer):
    x: FloatArr
    y: FloatArr

    def __init__(self, learning_rate: float = 0.01):
        self.learning_rate = learning_rate

    def __call__(self, parameters: FloatArr, gradients: FloatArr) -> FloatArr:
        return parameters - self.learning_rate*gradients

    def load_data(self, x_full: FloatArr, y_full:FloatArr):
        self.x = x_full
        self.y = y_full

    def get_batch(self):
        return self.x, self.y

class StochasticGradientDescend(Optimizer):
    batch: int
    batch_size: int
    x: FloatArr
    y: FloatArr

    def __init__(self, learning_rate: float = 0.01, batch_size: int = 50):
        self.learning_rate = learning_rate
        self.batch_size = batch_size
        self.batch = -1

    def __call__(self, parameters: FloatArr, gradients: FloatArr) -> FloatArr:
        return parameters - self.learning_rate*gradients

    def load_data(self, x_full: FloatArr, y_full:FloatArr):
        perm = np.random.permutation(len(x_full))
        self.x = x_full[perm]
        self.y = y_full[perm]

    def get_batch(self):
        if (self.batch + 1) * self.batch_size >= len(self.x[0]):
            self.batch = -1

        self.batch+=1
        start_idx, end_idx = self.batch * self.batch_size, (self.batch + 1) * self.batch_size
        return self.x[0:1, start_idx:end_idx], self.y[0:1, start_idx:end_idx]


